# -*implants -*-
"""

Functions for creating retinal implants

"""
import numpy as np
import logging

from pulse2percept import utils


SUPPORTED_IMPLANT_TYPES = ['epiretinal', 'subretinal']


class Electrode(object):

    def __init__(self, etype, radius, x_center, y_center, height=0, name=None):
        """Create an electrode on the retina

        This function creates a disk electrode of type `etype` and places it
        on the retina at location (`xs`, `ys`) in microns. The electrode has
        radius `radius` (microns) and sits a distance `height` away from the
        retinal surface.
        The coordinate system is anchored around the fovea at (0, 0).

        Parameters
        ----------
        etype : str
            Electrode type, {'epiretinal', 'subretinal'}
        radius : float
            The radius of the electrode (in microns).
        x_center : float
            The x coordinate of the electrode center (in microns) from the
            fovea.
        y_center : float
            The y location of the electrode (in microns) from the fovea
        height : float
            The height of the electrode from the retinal surface:

            - epiretinal array: distance to the ganglion layer
            - subretinal array: distance to the bipolar layer
        name : string
            Electrode name

        """
        assert radius >= 0
        assert height >= 0

        if etype.lower() not in SUPPORTED_IMPLANT_TYPES:
            e_s = "Acceptable values for `etype` are: "
            e_s += ", ".join(SUPPORTED_IMPLANT_TYPES) + "."
            raise ValueError(e_s)

        self.etype = etype.lower()
        self.radius = radius
        self.x_center = x_center
        self.y_center = y_center
        self.name = name
        self.height = height

    def __str__(self):
        info_s = "Electrode(%s, r=%.2f um, " % (self.etype, self.radius)
        info_s += "(x,y) = (%.2f, %.2f) um, " % (self.x_center, self.y_center)
        info_s += "h=%.2f um, n=%s" % (self.height, self.name)
        return info_s

    def get_height(self):
        """Returns the electrode-retina distance

        For epiretinal electrodes, this returns the distance to the ganglion
        cell layer.
        For subretinal electrodes, this returns the distance to the bipolar
        layer.
        """
        if self.etype == 'epiretinal':
            return self.h_ofl
        elif self.etype == 'subretinal':
            return self.h_inl
        else:
            raise ValueError("Unknown `etype`: " + self.etype)

    def set_height(self, height):
        """Sets the electrode-to-retina distance

        This function sets the electrode-to-retina distance according to
        `height`. For an epiretinal device, we calculate the distance to
        the ganglion cell layer (layer thickness depends on retinal location).
        For a subretinal device, we calculate the distance to the bipolar
        layer (layer thickness again depends on retinal location).

        Estimates of layer thickness based on:
        LoDuca et al. Am J. Ophthalmology 2011
        Thickness Mapping of Retinal Layers by Spectral Domain Optical
        Coherence Tomography
        Note that this is for normal retinal, so may overestimate thickness.
        Thickness from their paper (averaged across quadrants):
            0-600 um radius (from fovea):

            - Layer 1. (Nerve fiber layer) = 4
            - Layer 2. (Ganglion cell bodies + inner plexiform) = 56
            - Layer 3. (Bipolar bodies, inner nuclear layer) = 23

          600-1550 um radius:

            - Layer 1. 34
            - Layer 2. 87
            - Layer 3. 37.5

          1550-3000 um radius:
            - Layer 1. 45.5
            - Layer 2. 58.2
            - Layer 3. 30.75

        We place our ganglion axon surface on the inner side of the nerve fiber
        layer.
        We place our bipolar surface 1/2 way through the inner nuclear layer.
        So for an epiretinal array the bipolar layer is L1 + L2 + 0.5 * L3.

        """
        fovdist = np.sqrt(self.x_center ** 2 + self.y_center ** 2)
        if fovdist <= 600:
            # Layer thicknesses given for 0-600 um distance (from fovea)
            th_ofl = 4.0  # nerve fiber layer
            th_gc = 56.0  # ganglion cell bodies + inner nuclear layer
            th_bp = 23.0  # bipolar bodies + inner nuclear layer
        elif fovdist <= 1550:
            # Layer thicknesses given for 600-1550 um distance (from fovea)
            th_ofl = 34.0
            th_gc = 87.0
            th_bp = 37.5
        else:
            # Layer thicknesses given for 1550-3000 um distance (from fovea)
            th_ofl = 45.5
            th_gc = 58.2
            th_bp = 30.75
            if fovdist > 3000:
                e_s = "Distance to fovea=%.0f > 3000 um, " % fovdist
                e_s += "assuming same layer thicknesses as for 1550-3000 um "
                e_s += "distance."
                logging.getLogger(__name__).warning(e_s)

        if self.etype == 'epiretinal':
            # This is simply the electrode-retina distance
            self.h_ofl = height

            # All the way through the ganglion cell layer, inner plexiform
            # layer, and halfway through the inner nuclear layer
            self.h_inl = height + th_ofl + th_gc + 0.5 * th_bp
        elif self.etype == 'subretinal':
            # Starting from the outer plexiform layer, go halfway through the
            # inner nuclear layer
            self.h_inl = height + 0.5 * th_bp

            # Starting from the outer plexiform layer, all the way through the
            # inner nuclear layer, inner plexiform layer, and ganglion cell
            # layer
            self.h_ofl = height + th_bp + th_gc + th_ofl
        else:
            raise ValueError("Unknown `etype`: " + self.etype)
    height = property(get_height, set_height)

    def current_spread(self, xg, yg, layer, alpha=14000, n=1.69):
        """

        The current spread due to a current pulse through an electrode,
        reflecting the fall-off of the current as a function of distance from
        the electrode center. This can be calculated for any layer in the
        retina.
        Based on equation 2 in Nanduri et al [1].

        Parameters
        ----------
        xg : array
            x-coordinates of the retinal grid
        yg : array
            y-coordinates of the retinal grid
        layer: str
            Layer for which to calculate the current spread:

            - 'OFL': optic fiber layer, ganglion axons
            - 'INL': inner nuclear layer, containing the bipolars
        alpha : float
            A constant to do with the spatial fall-off.

        n : float
            A constant to do with the spatial fall-off (Default: 1.69, based
            on Ahuja et al. [2]  An In Vitro Model of a Retinal Prosthesis.
            Ashish K. Ahuja, Matthew R. Behrend, Masako Kuroda, Mark S.
            Humayun, and James D. Weiland (2008). IEEE Trans Biomed Eng 55.

        """
        r = np.sqrt((xg - self.x_center) ** 2 + (yg - self.y_center) ** 2)
        # current values on the retina due to array being above the retinal
        # surface
        if 'OFL' in layer:  # optic fiber layer, ganglion axons
            h = np.ones(r.shape) * self.h_ofl
            # actual distance from the electrode edge
            d = ((r - self.radius)**2 + self.h_ofl**2)**.5
        elif 'INL' in layer:  # inner nuclear layer, containing the bipolars
            h = np.ones(r.shape) * self.h_inl
            d = ((r - self.radius)**2 + self.h_inl**2)**.5
        else:
            s = "Layer %s not found. Acceptable values for `layer` are " \
                "'OFL' or 'INL'." % layer
            raise ValueError(s)
        cspread = (alpha / (alpha + h ** n))
        cspread[r > self.radius] = (alpha /
                                    (alpha + d[r > self.radius] ** n))

        return cspread

    def receptive_field(self, xg, yg, rftype='square', size=None):
        """An electrode's receptive field

        Parameters
        ----------
        xg : array_like
            Array of all x coordinates
        yg : array_like
            Array of all y coordinates
        rftype : {'square', 'gaussian'}
            The type of receptive field.
            - 'square': A simple square box receptive field with side length
                        `size`.
            - 'gaussian': A Gaussian receptive field where the weight drops off
                          as a function of distance from the electrode center.
                          The standard deviation of the Gaussian is `size`.
        size : float, optional
            Parameter describing the size of the receptive field. For square
            receptive fields, this corresponds to the side length of the
            square.
            For Gaussian receptive fields, this corresponds to the standard
            deviation of the Gaussian.
            Default: Twice the electrode radius.
        """
        if size is None:
            size = 2 * self.radius

        if rftype == 'square':
            # Create a map of the retina for each electrode
            # where it's 1 under the electrode, 0 elsewhere
            rf = np.zeros(xg.shape).astype(np.float32)
            ind = np.where((xg > self.x_center - (size / 2.0)) &
                           (xg < self.x_center + (size / 2.0)) &
                           (yg > self.y_center - (size / 2.0)) &
                           (yg < self.y_center + (size / 2.0)))
            rf[ind] = 1.0
        elif rftype == 'gaussian':
            # Create a map of the retina where the weight drops of as a
            # function of distance from the electrode center
            dist = (xg - self.x_center) ** 2 + (yg - self.y_center) ** 2
            rf = np.exp(-dist / (2 * size ** 2))
            rf /= np.sum(rf)
        else:
            e_s = "Acceptable values for `rftype` are 'square' or 'gaussian'"
            raise ValueError(e_s)

        return rf


class ElectrodeArray(object):

    def __init__(self, etype, radii, xs, ys, hs=0, names=None, eye='RE'):
        """Create an ElectrodeArray on the retina
        This function creates an electrode array of type `etype` and places it
        on the retina. Lists should specify, for each electrode, its size
        (`radii`), location on the retina (`xs` and `ys`), distance to the
        retina (height, `hs`), and a string identifier (`names`, optional).
        Array location should be given in microns, where the fovea is located
        at (0, 0).
        Single electrodes in the array can be addressed by index (integer)
        or name.
        Parameters
        ----------
        radii : array_like
            List of electrode radii.
        xs : array_like
            List of x-coordinates for the center of the electrodes (microns).
        ys : array_like
            List of y-coordinates for the center of the electrodes (microns).
        hs : float | array_like, optional, default: 0
            List of electrode heights (distance from the retinal surface).
        names : array_like, optional, default: None
            List of names (string identifiers) for each eletrode.
        eye : {'LE', 'RE'}, optional, default: 'RE'
            Eye in which array is implanted.

        Examples
        --------
        A single epiretinal electrode called 'A1', with radius 100um, sitting
        at retinal location (0, 0), 10um away from the retina:
        >>> from pulse2percept import implants
        >>> implant0 = implants.ElectrodeArray('epiretinal', 100, 0, 0, hs=10,
        ...                                    names='A1')

        Get access to the electrode with name 'A1' in the first array:
        >>> my_electrode = implant0['A1']

        An array with two electrodes of size 100um, one sitting at
        (-100, -100), the other sitting at (0, 0), with 0 distance from the
        retina, of type 'subretinal':
        >>> implant1 = implants.ElectrodeArray('subretinal', [100, 100],
        ...                                    [-100, 0], [-100, 0], hs=[0, 0])
        """
        self.etype = etype
        self.eye = eye
        self.electrodes = []
        self.num_electrodes = 0
        self.add_electrodes(radii, xs, ys, hs, names)

    def __str__(self):
        return "ElectrodeArray(%s, num_electrodes=%d)" % (self.etype,
                                                          self.num_electrodes)

    def add_electrode(self, electrode):
        """Adds an electrode to an ElectrodeArray object
        This function adds a single electrode to an existing ElectrodeArray
        object. The electrode must have the same type as the array
        (see implants.SUPPORTED_IMPLANT_TYPES).
        Parameters
        ----------
        electrode : implants.Electrode
            An electrode object specifying type, size, and location of the
            electrode on the retina.
        """
        if not isinstance(electrode, Electrode):
            raise TypeError("`electrode` must be of type retina.Electrode.")

        if electrode.etype != self.etype:
            e_s = "Added electrode must be of same type as the existing"
            e_s = "array (%s)." % self.etype
            raise ValueError(e_s)

        self.num_electrodes += 1
        self.electrodes.append(electrode)

    def add_electrodes(self, radii, xs, ys, hs=0, names=None):
        """Adds electrodes to an ElectrodeArray object
        This function adds one or more electrodes to an existing ElectrodeArray
        object. Lists should specify, for each electrode to be added, the size
        (`radii`), location on the retina (`xs` and `ys`), distance to the
        retina (height, `hs`), and a string identifier (`names`, optional).
        Array location should be given in microns, where the fovea is located
        at (0, 0).
        Single electrodes in the array can be addressed by index (integer)
        or name.

        Parameters
        ----------
        radii : array_like
            List of electrode radii.
        xs : array_like
            List of x-coordinates for the center of the electrodes (microns).
        ys : array_like
            List of y-coordinates for the center of the electrodes (microns).
        hs : float | array_like, optional, default: 0
            List of electrode heights (distance from the retinal surface).
        names : array_like, optional, default: None
            List of names (string identifiers) for each eletrode.

        Examples
        --------
        Adding a single electrode of radius 50um sitting at (0, 0) to an
        existing ElectrodeArray object:
        >>> implant = ElectrodeArray('epiretinal', 100, 100, 100)
        >>> implant.add_electrodes(50, 0, 0)
        """
        # Make it so the method can accept either floats, lists, or
        # numpy arrays, and `zip` works regardless.
        radii = np.array([radii], dtype=np.float32).flatten()
        xs = np.array([xs], dtype=np.float32).flatten()
        ys = np.array([ys], dtype=np.float32).flatten()
        names = np.array([names], dtype=np.str).flatten()

        if isinstance(hs, list):
            hs = np.array(hs).flatten()
        else:
            # All electrodes have the same height
            hs = np.ones_like(radii) * hs

        assert radii.size == xs.size == ys.size == hs.size

        if names.size != radii.size:
            # If not every electrode has a name, replace with None's
            names = np.array([None] * radii.size)

        for r, x, y, h, n in zip(radii, xs, ys, hs, names):
            self.add_electrode(Electrode(self.etype, r, x, y, h, n))

    def __iter__(self):
        return iter(self.electrodes)

    def __getitem__(self, item):
        """Return the electrode specified by `item`
        Parameters
        ----------
        item : int|string
            If `item` is an integer, returns the `item`-th electrode in the
            array. If `item` is a string, returns the electrode with string
            identifier `item`.
        """
        try:
            # Is `item` an integer?
            return self.electrodes[item]
        except (IndexError, TypeError):
            # If `item` is a valid string identifier, return valid index.
            # Else return None
            try:
                return self.electrodes[self.get_index(item)]
            except (IndexError, TypeError):
                return None

    def get_index(self, name):
        """Returns the index of an electrode called `name`
        This function searches the electrode array for an electrode with
        string identifier `name`. If found, the index of that electrode is
        returned, else None.
        Parameters
        ----------
        name : str
            An electrode name (string identifier).
        Returns
        -------
        A valid electrode index or None.
        """
        # Is `name` a valid electrode name?
        # Iterate through electrodes to find a matching name. Shuffle list
        # to reduce time complexity of average lookup.
        for idx, el in utils.traverse_randomly(enumerate(self.electrodes)):
            if el.name == name:
                return idx

        # Worst case O(n): name could not be found.
        return None

    def get_eye(self):
        return self._eye

    def set_eye(self, eye):
        if eye.lower() in ['r', 're', 'right']:
            self._eye = 'RE'
        elif eye.lower() in ['l', 'le', 'left']:
            self._eye = 'LE'
        else:
            raise ValueError("Unknown eye '%s'. Choose from 'LE', 'RE'.")

    eye = property(get_eye, set_eye)


class ArgusI(ElectrodeArray):

    def __init__(self, x_center=0, y_center=0, h=0, rot=0, eye='RE',
                 use_legacy_names=False):
        """Create an ArgusI array on the retina
        This function creates an ArgusI array and places it on the retina
        such that the center of the array is located at
        [`x_center`, `y_center`] (microns) and the array is rotated by
        rotation angle `rot` (radians).
        The array is oriented in the visual field as shown in Fig. 1 of
        Horsager et al. (2009); that is, if placed in (0,0), the top two
        rows will lie in the lower retina (upper visual field):
        .. raw:: html
          <pre>
            y       A1 B1 C1 D1                     260 520 260 520
            ^       A2 B2 C2 D2   where electrode   520 260 520 260
            |       A3 B3 C3 D3   diameters are:    260 520 260 520
            -->x    A4 B4 C4 D4                     520 260 520 260
          </pre>
        Electrode order is: A1, B1, C1, D1, A2, B2, ..., D4.
        If `use_legacy_names` is True, electrode order is: L6, L2, M8, M4, ...
        An electrode can be addressed by index (integer) or name.

        Parameters
        ----------
        x_center : float, optional, default: 0
            x coordinate of the array center (um)
        y_center : float, optional, default: 0
            y coordinate of the array center (um)
        h : float || array_like, optional, default: 0
            Distance of the array to the retinal surface (um). Either a list
            with 16 entries or a scalar.
        rot : float, optional, default: 0
            Rotation angle of the array (rad). Positive values denote
            counter-clock-wise (CCW) rotations in the retinal coordinate
            system.
        eye : {'LE', 'RE'}, optional, default: 'RE'
            Eye in which array is implanted.

        Examples
        --------
        Create an ArgusI array centered on the fovea, at 100um distance from
        the retina:
        >>> from pulse2percept import implants
        >>> argus = implants.ArgusI(x_center=0, y_center=0, h=100, rot=0)

        Get access to electrode 'B1':
        >>> my_electrode = argus['B1']
        """
        # Alternating electrode sizes, arranged in checkerboard pattern
        r_arr = np.array([260, 520, 260, 520]) / 2.0
        r_arr = np.concatenate((r_arr, r_arr[::-1], r_arr, r_arr[::-1]),
                               axis=0)

        # Set left/right eye
        self.eye = eye

        # In older papers, Argus I electrodes go by L and M
        self.old_names = names = ['L6', 'L2', 'M8', 'M4',
                                  'L5', 'L1', 'M7', 'M3',
                                  'L8', 'L4', 'M6', 'M2',
                                  'L7', 'L3', 'M5', 'M1']
        # In newer papers, they go by A-D: A1, B1, C1, D1, A1, B2, ..., D4
        # Shortcut: Use `chr` to go from int to char
        self.new_names = [chr(i) + str(j) for j in range(1, 5)
                          for i in range(65, 69)]

        if use_legacy_names:
            names = self.old_names
        else:
            names = self.new_names

        if isinstance(h, list):
            h_arr = np.array(h).flatten()
            if h_arr.size != len(r_arr):
                e_s = "If `h` is a list, it must have 16 entries."
                raise ValueError(e_s)
        else:
            # All electrodes have the same height
            h_arr = np.ones_like(r_arr) * h

        # Equally spaced electrodes: n_rows x n_cols = 16
        e_spacing = 800  # um
        n_cols = 4  # number of electrodes horizontally (same vertically)
        x_arr = np.arange(n_cols) * e_spacing - (n_cols / 2 - 0.5) * e_spacing
        if self.eye == 'LE':
            # Left eye: Need to invert x coordinates and rotation angle
            x_arr = x_arr[::-1]
        x_arr, y_arr = np.meshgrid(x_arr, x_arr, sparse=False)

        # Rotation matrix
        R = np.array([np.cos(rot), -np.sin(rot),
                      np.sin(rot), np.cos(rot)]).reshape((2, 2))

        # Set the x, y location of the tack
        if self.eye == 'RE':
            self.tack = np.matmul(R, [-(n_cols / 2 + 0.5) * e_spacing, 0])
        else:
            self.tack = np.matmul(R, [(n_cols / 2 + 0.5) * e_spacing, 0])
        self.tack = tuple(self.tack + [x_center, y_center])

        # Rotate the array
        xy = np.vstack((x_arr.flatten(), y_arr.flatten()))
        xy = np.matmul(R, xy)
        x_arr = xy[0, :]
        y_arr = xy[1, :]

        # Apply offset
        x_arr += x_center
        y_arr += y_center

        self.etype = 'epiretinal'
        self.num_electrodes = 0
        self.electrodes = []
        for r, x, y, h, n in zip(r_arr, x_arr, y_arr, h_arr, names):
            self.add_electrode(Electrode(self.etype, r, x, y, h, n))

    def __str__(self):
        return "ArgusI(%s, num_electrodes=%d)" % (self.etype,
                                                  self.num_electrodes)

    def get_old_name(self, new_name):
        """Look up the legacy name of a standard-named Argus I electrode"""
        return self.old_names[self.new_names.index(new_name)]

    def get_new_name(self, old_name):
        """Look up the standard name of a legacy-named Argus I electrode"""
        return self.new_names[self.old_names.index(old_name)]


class ArgusII(ElectrodeArray):

    def __init__(self, x_center=0, y_center=0, h=0, rot=0, eye='RE'):
        """Create an ArgusII array on the retina
        This function creates an ArgusII array and places it on the retina
        such that the center of the array is located at
        [`x_center`, `y_center`] (microns) and the array is rotated by
        rotation angle `rot` (radians).
        The array is oriented upright in the visual field, such that an
        array with center (0,0) has the top three rows lie in the lower
        retina (upper visual field), as shown below:
        .. raw:: html
          <pre>
                    A1 A2 A3 A4 A5 A6 A7 A8 A9 A10
            y       B1 B2 B3 B4 B5 B6 B7 B8 B9 B10
            ^       C1 C2 C3 C4 C5 C6 C7 C8 C9 C10
            |       D1 D2 D3 D4 D5 D6 D7 D8 D9 D10
            -->x    E1 E2 E3 E4 E5 E6 E7 E8 E9 E10
                    F1 F2 F3 F4 F5 F6 F7 F8 F9 F10
          </pre>
        Electrode order is: A1, A2, ..., A10, B1, B2, ..., F10.
        An electrode can be addressed by index (integer) or name.

        Parameters
        ----------
        x_center : float
            x coordinate of the array center (um)
        y_center : float
            y coordinate of the array center (um)
        h : float || array_like
            Distance of the array to the retinal surface (um). Either a list
            with 60 entries or a scalar.
        rot : float
            Rotation angle of the array (rad). Positive values denote
            counter-clock-wise (CCW) rotations in the retinal coordinate
            system.
        eye : {'LE', 'RE'}, optional, default: 'RE'
            Eye in which array is implanted.

        Examples
        --------
        Create an ArgusII array centered on the fovea, at 100um distance from
        the retina:
        >>> from pulse2percept import implants
        >>> argus = implants.ArgusII(x_center=0, y_center=0, h=100, rot=0)

        Get access to electrode 'E7':
        >>> my_electrode = argus['E7']
        """
        # Electrodes are 200um in diameter
        r_arr = np.ones(60) * 100.0

        # Set left/right eye
        self.eye = eye

        # Standard ArgusII names: A1, A2, ..., A10, B1, ..., F10
        names = [chr(i) + str(j) for i in range(65, 71) for j in range(1, 11)]

        if isinstance(h, list):
            h_arr = np.array(h).flatten()
            if h_arr.size != len(r_arr):
                e_s = "If `h` is a list, it must have 60 entries."
                raise ValueError(e_s)
        else:
            # All electrodes have the same height
            h_arr = np.ones_like(r_arr) * h

        # Equally spaced electrodes: n_rows x n_cols = 60
        e_spacing = 525  # um
        n_cols = 10  # number of electrodes horizontally
        n_rows = 6  # number of electrodes vertically
        x_arr = np.arange(n_cols) * e_spacing - (n_cols / 2 - 0.5) * e_spacing
        if self.eye == 'LE':
            # Left eye: Need to invert x coordinates and rotation angle
            x_arr = x_arr[::-1]
        y_arr = np.arange(n_rows) * e_spacing - (n_rows / 2 - 0.5) * e_spacing
        x_arr, y_arr = np.meshgrid(x_arr, y_arr, sparse=False)

        # Rotation matrix
        rotmat = np.array([np.cos(rot), -np.sin(rot),
                           np.sin(rot), np.cos(rot)]).reshape((2, 2))

        # Set the x, y location of the tack
        if self.eye == 'RE':
            self.tack = np.matmul(rotmat, [-(n_cols / 2 + 0.5) * e_spacing, 0])
        else:
            self.tack = np.matmul(rotmat, [(n_cols / 2 + 0.5) * e_spacing, 0])
        self.tack = tuple(self.tack + [x_center, y_center])

        # Rotate the array
        xy = np.vstack((x_arr.flatten(), y_arr.flatten()))
        xy = np.matmul(rotmat, xy)
        x_arr = xy[0, :]
        y_arr = xy[1, :]

        # Apply offset
        x_arr += x_center
        y_arr += y_center

        self.etype = 'epiretinal'
        self.num_electrodes = 0
        self.electrodes = []
        for r, x, y, h, n in zip(r_arr, x_arr, y_arr, h_arr, names):
            self.add_electrode(Electrode(self.etype, r, x, y, h, n))

    def __str__(self):
        return "ArgusII(%s, num_electrodes=%d)" % (self.etype,
                                                   self.num_electrodes)



import numpy as np
import logging
import six

from pulse2percept import (utils, retina, implants, stimuli)


class Simulation(object):

    def __init__(self, implant, engine='joblib', scheduler='threading',
                 use_jit=True, n_jobs=-1):
        """Generates a simulation framework

        Parameters
        ----------
        implant : implants.ElectrodeArray
            An implants.ElectrodeArray object that describes the implant.
        engine : str, optional, default: 'joblib'
            Which computational back end to use:
            - 'serial': Single-core computation
            - 'joblib': Parallelization via joblib (requires `pip install
                        joblib`)
            - 'dask': Parallelization via dask (requires `pip install dask`).
                      Dask backend can be specified via `threading`.
        scheduler : str, optional, default: 'threading'
            Which scheduler to use (irrelevant for 'serial' engine):
            - 'threading': a scheduler backed by a thread pool
            - 'multiprocessing': a scheduler backed by a process pool
        use_jit : bool, optional, default: True
            Whether to use just-in-time (JIT) compilation to speed up
            computation.
        n_jobs : int, optional, default: -1
            Number of cores (threads) to run the model on in parallel.
            Specify -1 to use as many cores as available.
        """
        if not isinstance(implant, implants.ElectrodeArray):
            e_s = "`implant` must be of type implants.ElectrodeArray"
            raise TypeError(e_s)

        self.implant = implant
        self.engine = engine
        self.scheduler = scheduler
        self.use_jit = use_jit
        self.n_jobs = n_jobs

        # Optic fiber layer (OFL): After calling `set_optic_fiber_layer`, this
        # variable will contain a `retina.Grid` object.
        self.ofl = None

        # Ganglion cell layer (GCL): After calling `set_ganglion_cell_layer`,
        # this variable will contain a `retina.TemporalModel` object.
        self.gcl = None

    def set_optic_fiber_layer(self, sampling=100, x_range=None, y_range=None,
                              n_axons=501, phi_range=(-180.0, 180.0),
                              n_rho=801, rho_range=(4.0, 45.0),
                              loc_od=(15.5, 1.5),
                              sensitivity_rule='decay', decay_const=1.0,
                              contribution_rule='max', powermean_exp=None,
                              datapath='.', save_data=True):
        """Sets parameters of the optic fiber layer (OFL)

        Parameters
        ----------
        sampling : float, optional, default: 100 microns
            Microns per grid cell.
        x_range : (xlo, xhi)|None, default: None
            Lower and upper bound of the retinal grid (microns) in horizontal
            dimension. Either a tuple (xlo, xhi) or None. If None, the
            generated grid will be just big enough to fit the implant.
        y_range : (ylo, yhi)|None, default: None
            Lower and upper bound of the retinal grid (microns) in vertical
            dimension. Either a tuple (ylo, yhi) or None. If None, the
            generated grid will be just big enough to fit the implant.
        n_axons : int, optional, default: 501
            The number of axons to generate. Their start orientations `phi0`
            (in modified polar coordinates) will be sampled uniformly from
            `phi_range`.
        phi_range : (lophi, hiphi), optional, default: (-180, 180)
            Range of angular positions of axon fibers at their starting points
            (polar coordinates, degrees) to be sampled uniformly with `n_axons`
            samples. Must be within [-180, 180].
        n_rho: int, optional, default: 801
            Number of sampling points along the radial axis(polar coordinates).
        rho_range: (rho_min, rho_max), optional, default: (4.0, 45.0)
            Lower and upper bounds for the radial position values(polar
            coordinates).
        loc_od: (x_od, y_od), optional, default: (15.5 1.5)
            Location of the center of the optic disc (x, y) in Cartesian
            coordinates. In a right (left) eye, we should have x > 0 (x < 0).
        sensitivity_rule : {'decay', 'Jeng2011'}, optional, default: 'decay'
            This rule specifies how the activation of the axon differs as a
            function of distance from the soma. The following options are
            available:
            - 'decay':
                Axon sensitivity decays exponentially with distance. Specify
                `decay_const` to change the steepness of the fall-off with
                distance.
            - 'Jeng2011':
                Axon sensitivity peaks near the sodium band (50um from the
                soma), then plateaus on the distal axon at roughly half of the
                peak sensitivity (see Fig. 2 in [1]_).
        contribution_rule : {'max', 'sum', 'mean'}, optional, default: 'max'
            This rule specifies how the activation thresholds across all axon
            segments are combined to determine the contribution of the axon to
            the current spread. The following options are available:
            - 'max':
                The axon's contribution to the current spread is equal to the
                max. sensitivity across all axon segments.
            - 'sum':
                The axon's contribution to the current spread is equal to the
                sum sensitivity across all axon segments.
            - 'mean':
                The axon's contribution to the current spread is equal to the
                mean sensitivity across all axon segments. Specify
                `powermean_exp` to change the exponent of the generalized
                (power) mean, calculated as np.mean(x ** powermean_exp) **
                (1.0 / powermean_exp). Default is 1, which is equal to the
                arithmetic mean.
        decay_const : float, optional, default: 2.0
            When `sensitivity_rule` is set to 'decay', specifies the decay
            constant of the exponential fall-off.
        powermean_exp : float, optional, default: 1.0
            When `sensitivity_rule` is set to 'mean', specifies the exponent of
            the generalized (power) mean function. The power mean is calculated
            as np.mean(x ** powermean_exp) ** (1.0 / powermean_exp).
        datapath : str, default: current directory
            Relative path where to look for existing retina files, and where to
            store new retina files.
        save_data : bool, default: True
            Flag whether to save the data to a new retina file (True) or not
            (False). The file name is automatically generated from all
            specified input arguments.

        Notes
        -----
        .. [1] J. Jeng, S. Tang, A. Molnar, N. J. Desai, and S. I. Fried, "The
           sodium channel band shapes the response to electric stimulation in
           retinal ganglion cells", J Neural Eng 8 (036022), 2011.
        """
        # For auto-generated grids:
        round_to = 500  # round to nearest (microns)
        cspread = 500  # add padding for current spread (microns)

        if x_range is None:
            # No x ranges given: generate automatically to fit the implant
            xs = [a.x_center for a in self.implant]
            xlo = np.floor((np.min(xs) - cspread) / round_to) * round_to
            xhi = np.ceil((np.max(xs) + cspread) / round_to) * round_to
        elif isinstance(x_range, (int, float)):
            xlo = x_range
            xhi = x_range
        elif isinstance(x_range, (list, tuple, np.ndarray)):
            if len(x_range) != 2 or x_range[1] < x_range[0]:
                e_s = "x_range must be a tuple (xlo, xhi) where xlo <= xhi."
                raise ValueError(e_s)
            xlo = x_range[0]
            xhi = x_range[1]
        else:
            raise ValueError("x_range must be a tuple (xlo, xhi) or None.")

        if y_range is None:
            # No y ranges given: generate automatically to fit the implant
            ys = [a.y_center for a in self.implant]
            ylo = np.floor((np.min(ys) - cspread) / round_to) * round_to
            yhi = np.ceil((np.max(ys) + cspread) / round_to) * round_to
        elif isinstance(y_range, (int, float)):
            ylo = y_range
            yhi = y_range
        elif isinstance(y_range, (list, tuple, np.ndarray)):
            if len(y_range) != 2 or y_range[1] < y_range[0]:
                e_s = "y_range must be a tuple (ylo, yhi) where ylo <= yhi."
                raise ValueError(e_s)
            ylo = y_range[0]
            yhi = y_range[1]
        else:
            raise ValueError("y_range must be a tuple (ylo, yhi) or None.")

        # Generate the grid from the above specs
        self.ofl = retina.Grid(x_range=(xlo, xhi), y_range=(ylo, yhi),
                               eye=self.implant.eye, sampling=sampling,
                               n_axons=n_axons, phi_range=phi_range,
                               n_rho=n_rho, rho_range=rho_range,
                               loc_od=loc_od,
                               sensitivity_rule=sensitivity_rule,
                               contribution_rule=contribution_rule,
                               decay_const=decay_const,
                               powermean_exp=powermean_exp,
                               datapath=datapath, save_data=save_data,
                               engine=self.engine, scheduler=self.scheduler,
                               n_jobs=self.n_jobs)

    def set_ganglion_cell_layer(self, model, **kwargs):
        """Sets parameters of the ganglion cell layer (GCL)

        Select from pre-existing ganglion cell models or specify a custom one.

        Parameters
        ----------
        model : str|retina.BaseModel
            A custom ganglion cell model can be specified by passing an
            instance of type `retina.BaseModel`. Else select from pre-existing
            models:

            - 'latest':
                The latest temporal model for epiretinal and subretinal
                arrays (experimental).

                Additional keyword arguments:

                - tau_gcl : float, optional, default: 45.25 ms
                    Time decay constant for the fast leaky integrater of the
                    ganglionc ell layer.
                - tau_inl : float, optional, default: 18 ms
                    Time decay constant for the fast leaky integrater of the
                    inner nuclear layer (INL). It has been shown that even
                    epiretinal arrays can activate bipolar cells (in the INL),
                    which in turn influence GCL activity.
                - tau_ca : float, optional, default: 45.25 ms
                    Time decay constant for the charge accumulation.
                - scale_ca : float, optional, default: 42.1
                    Scaling factor applied to charge accumulation (used to be
                    called epsilon).
                - tau_slow : float, optional, default: 26.25 ms
                    Time decay constant for the slow leaky integrator.
                - scale_slow : float, optional, default: 1150.0
                    Scaling factor applied to the output of the cascade, to
                    make output values interpretable brightness values >= 0.
                - lweight : float, optional, default: 0.636
                    Relative weight applied to responses from bipolar cells
                    (weight of ganglion cells is 1).
                - aweight : float, optional, default: 0.5
                    Relative weight applied to anodic charges (weight of
                    cathodic charges is 1).
                - slope : float, optional, default: 3.0
                    Slope of the logistic function in the stationary
                    nonlinearity stage.
                - shift : float, optional, default: 15.0
                    Shift of the logistic function in the stationary
                    nonlinearity stage.

            - 'Nanduri2012':
                A model of temporal sensitivity as described in [1]_.

                Additional keyword arguments
                ----------------------------
                tsample : float, optional, default:
                tau1 : float, optional, default: 0.42 / 1000 (seconds)
                    Time decay constant for the fast leaky integrater of
                    the ganglion cell layer (GCL).
                tau2 : float, optional, default: 45.25 / 1000 (seconds)
                    Time decay constant for the charge accumulation, has
                    values between 38 - 57 ms.
                tau3 : float, optional, default: 26.25 / 1000 (seconds)
                    Time decay constant for the slow leaky integrator.
                    Default: 26.25 / 1000 s.
                eps : float, optional, default: 8.73
                    Scaling factor applied to charge accumulation (used to
                    be called epsilon).
                asymptote : float, optional, default: 14.0
                    Asymptote of the logistic function used in the
                    stationary nonlinearity stage.
                slope : float, optional, default: 3.0
                    Slope of the logistic function in the stationary
                    nonlinearity stage.
                shift : float, optional, default: 16.0
                    Shift of the logistic function in the stationary
                    nonlinearity stage.

            - 'Horsager2009':
                A model of temporal sensitivity as described in [2]_.

                Parameters
                ----------
                tsample : float, optional, default: 0.005 / 1000 seconds
                    Sampling time step (seconds).
                tau1 : float, optional, default: 0.42 / 1000 seconds
                    Time decay constant for the fast leaky integrater of the
                    ganglion cell layer (GCL).
                tau2 : float, optional, default: 45.25 / 1000 seconds
                    Time decay constant for the charge accumulation, has values
                    between 38 - 57 ms.
                tau3 : float, optional, default: 26.25 / 1000 seconds
                    Time decay constant for the slow leaky integrator.
                    Default: 26.25 / 1000 s.
                eps : float, optional, default: 8.73
                    Scaling factor applied to charge accumulation (used to be
                    called epsilon).
                beta : float, optional, default: 3.43
                    Power nonlinearity applied after half-rectification. The
                    original model used two different values, depending on
                    whether an experiment is at threshold (`beta`=3.43) or
                    above threshold (`beta`=0.83).

        Notes
        -----
        .. [1] D. Nanduri, I. Fine, A. Horsager, G. M. Boynton, M. S. Humayun,
               R. Greenberg, J. D. Weiland, "Frequency and Amplitude Modulation
               Have Different Effects on the Percepts Elicited by Retinal
               Stimulation", Investigative Ophthalmology & Visual Science 53,
               205-214, 2012.
        .. [2] A. Horsager, S. H. Greenwald, J. D. Weiland, M. S. Humayun,
               R. J. Greenberg, M. J. McMahon, G. M. Boynton, and I. Fine,
               "Predicting visual sensitivity in retinal prosthesis patients",
               Investigative Ophthalmology & Visual Science, 50(4):1483, 2009.



        """
        model_not_found = False
        if isinstance(model, six.string_types):
            # If `model` is a string, choose from existing models
            if model.lower() == 'latest':
                logging.getLogger(__name__).debug("Setting up latest model.")
                self.gcl = retina.TemporalModel(**kwargs)
            elif model.lower() in ['nanduri', 'nanduri2012']:
                logging.getLogger(__name__).debug("Setting up Nanduri (2012) "
                                                  "model.")
                self.gcl = retina.Nanduri2012(**kwargs)
            elif model.lower() in ['horsager', 'horsager2009']:
                logging.getLogger(__name__).debug("Setting up Horsager "
                                                  "(2009) model.")
                self.gcl = retina.Horsager2009(**kwargs)
            else:
                model_not_found = True
        elif isinstance(model, retina.BaseModel):
            # If `model` is not a string, must be of type BaseModel
            debug_str = "Setting up %s." % model.__module__
            logging.getLogger(__name__).debug(debug_str)
            self.gcl = model
        else:
            model_not_found = True

        if model_not_found:
            err_str = "Model '%s' not found. Choose from: " % model
            err_str += ", ".join(retina.SUPPORTED_TEMPORAL_MODELS)
            err_str += " or provide your own retina.BaseModel instance."
            raise ValueError(err_str)

    def _set_layers(self):
        """Sets up all layers whose setters have not been called by the user

        This function makes sure all necessary parts of the simulation are
        initialized before transforming stimuli to percepts.
        Layers not initialized by the user will simply be initialized with
        default argument values.
        """
        if self.ofl is None:
            self.set_optic_fiber_layer()
        if self.gcl is None:
            self.set_ganglion_cell_layer('latest')

    def pulse2percept(self, stim, t_percept=None, tol=0.05,
                      layers=['OFL', 'GCL', 'INL']):
        """Transforms an input stimulus to a percept

        Parameters
        ----------
        stim : utils.TimeSeries|list|dict
            There are several ways to specify an input stimulus:

            - For a single-electrode array, pass a single pulse train; i.e.,
              a single utils.TimeSeries object.
            - For a multi-electrode array, pass a list of pulse trains; i.e.,
              one pulse train per electrode.
            - For a multi-electrode array, specify all electrodes that should
              receive non-zero pulse trains by name.
        t_percept : float, optional, default: inherit from `stim` object
            The desired time sampling of the output (seconds).
        tol : float, optional, default: 0.05
            Ignore pixels whose effective current is smaller than a fraction
            `tol` of the max value.
        layers : list, optional, default: ['OFL', 'GCL', 'INL']
            A list of retina layers to simulate (order does not matter):
            - 'OFL': Includes the optic fiber layer in the simulation.
                     If omitted, the tissue activation map will not account
                     for axon streaks.
            - 'GCL': Includes the ganglion cell layer in the simulation.
            - 'INL': Includes the inner nuclear layer in the simulation.
                     If omitted, bipolar cell activity does not contribute
                     to ganglion cell activity.

        Returns
        -------
        A utils.TimeSeries object whose data container comprises the predicted
        brightness over time at each retinal location (x, y), with the last
        dimension of the container representing time (t).

        Examples
        --------
        Simulate a single-electrode array:

        >>> import pulse2percept as p2p
        >>> implant = p2p.implants.ElectrodeArray('subretinal', 0, 0, 0)
        >>> stim = p2p.stimuli.PulseTrain(tsample=5e-6, freq=50, amp=20)
        >>> sim = p2p.Simulation(implant)
        >>> percept = sim.pulse2percept(stim)  # doctest: +SKIP

        Simulate an Argus I array centered on the fovea, where a single
        electrode is being stimulated ('C3'):

        >>> import pulse2percept as p2p
        >>> implant = p2p.implants.ArgusI()
        >>> stim = {'C3': stimuli.PulseTrain(tsample=5e-6, freq=50,
        ...                                              amp=20)}
        >>> sim = p2p.Simulation(implant)
        >>> resp = sim.pulse2percept(stim, implant)  # doctest: +SKIP
        """
        logging.getLogger(__name__).info("Starting pulse2percept...")

        # Get a flattened, all-uppercase list of layers
        layers = np.array([layers]).flatten()
        layers = np.array([l.upper() for l in layers])

        # Make sure all specified layers exist
        not_supported = np.array([l not in retina.SUPPORTED_LAYERS
                                  for l in layers], dtype=bool)
        if any(not_supported):
            msg = ', '.join(layers[not_supported])
            msg = "Specified layer %s not supported. " % msg
            msg += "Choose from %s." % ', '.join(retina.SUPPORTED_LAYERS)
            raise ValueError(msg)

        # Set up all layers that haven't been set up yet
        self._set_layers()

        # Parse `stim` (either single pulse train or a list/dict of pulse
        # trains), and generate a list of pulse trains, one for each electrode
        pt_list = stimuli.parse_pulse_trains(stim, self.implant)
        pt_data = [pt.data for pt in pt_list]

        if not np.allclose([p.tsample for p in pt_list], self.gcl.tsample):
            e_s = "For now, all pulse trains must have the same sampling "
            e_s += "time step as the ganglion cell layer. In the future, "
            e_s += "this requirement might be relaxed."
            raise ValueError(e_s)

        # Tissue activation maps: If OFL is simulated, includes axon streaks.
        if 'OFL' in layers:
            ecs, _ = self.ofl.electrode_ecs(self.implant)
        else:
            _, ecs = self.ofl.electrode_ecs(self.implant)

        # Calculate the max of every current spread map
        lmax = np.zeros((2, ecs.shape[-1]))
        if 'INL' in layers:
            lmax[0, :] = ecs[:, :, 0, :].max(axis=(0, 1))
        if ('GCL' or 'OFL') in layers:
            lmax[1, :] = ecs[:, :, 1, :].max(axis=(0, 1))

        # `ecs_list` is a pixel by `n` list where `n` is the number of layers
        # being simulated. Each value in `ecs_list` is the current contributed
        # by each electrode for that spatial location
        ecs_list = []
        idx_list = []
        for xx in range(self.ofl.gridx.shape[1]):
            for yy in range(self.ofl.gridx.shape[0]):
                # If any of the used current spread maps at [yy, xx] are above
                # tolerance, we need to process that pixel
                process_pixel = False
                if 'INL' in layers:
                    # For this pixel: Check if the ecs in any layer is large
                    # enough compared to the max across pixels within the layer
                    process_pixel |= np.any(ecs[yy, xx, 0, :] >=
                                            tol * lmax[0, :])
                if ('GCL' or 'OFL') in layers:
                    process_pixel |= np.any(ecs[yy, xx, 1, :] >=
                                            tol * lmax[1, :])

                if process_pixel:
                    ecs_list.append(ecs[yy, xx])
                    idx_list.append([yy, xx])

        s_info = "tol=%.1f%%, %d/%d px selected" % (tol * 100, len(ecs_list),
                                                    np.prod(ecs.shape[:2]))
        logging.getLogger(__name__).info(s_info)

        sr_list = utils.parfor(self.gcl.model_cascade,
                               ecs_list, n_jobs=self.n_jobs,
                               engine=self.engine, scheduler=self.scheduler,
                               func_args=[pt_data, layers, self.use_jit])
        bm = np.zeros(self.ofl.gridx.shape +
                      (sr_list[0].data.shape[-1], ))
        idxer = tuple(np.array(idx_list)[:, i] for i in range(2))
        bm[idxer] = [sr.data for sr in sr_list]
        percept = utils.TimeSeries(sr_list[0].tsample, bm)

        # It is possible to specify an additional sampling rate for the
        # percept: If different from the input sampling rate, need to resample.
        if t_percept != percept.tsample:
            percept = percept.resample(t_percept)

        logging.getLogger(__name__).info("Done.")

        return percept

    def plot_fundus(self, stim=None, ax=None, n_axons=100, upside_down=False,
                    annotate=True):
        """Plot the implant on the retinal surface akin to a fundus photopgraph
        This function plots an electrode array on top of the axon streak map
        of the retina, akin to a fundus photograph. A blue rectangle highlights
        the area of the retinal surface that is being simulated.
        If `stim` is passed, activated electrodes will be highlighted.
        Parameters
        ----------
        stim : utils.TimeSeries|list|dict, optional
            An input stimulus, as passed to ``p2p.pulse2percept``. If given,
            activated electrodes will be highlighted in the plot.
            Default: None
        ax : matplotlib.axes._subplots.AxesSubplot, optional
            A Matplotlib axes object. If None given, a new one will be created.
            Default: None
        n_axons : int, optional, default: 100
            Number of axons to plot.
        upside_down : bool, optional, default: False
            Flag whether to plot the retina upside-down, such that the upper
            half of the plot corresponds to the upper visual field. In general,
            inferior retina == upper visual field (and superior == lower).
        annotate : bool, optional, default: True
            Flag whether to annotate the four retinal quadrants
            (inferior/superior x temporal/nasal).

        Returns
        -------
        Returns a handle to the created figure (`fig`) and axes element (`ax`).
        """
        if self.ofl is not None:
            # Optic fiber layer already set via  ``set_optic_fiber_layer``
            phi_range = self.ofl.phi_range
            n_rho = self.ofl.n_rho
            rho_range = self.ofl.rho_range
            loc_od = self.ofl.loc_od
            x_range = self.ofl.x_range
            y_range = self.ofl.y_range
            plot_patch = True
        else:
            msg = "Optic fiber layer not set (call ``set_optic_fiber_layer``)."
            msg += " Using default values."
            logging.getLogger(__name__).info(msg)
            phi_range = (-180.0, 180.0)
            n_rho = 801
            rho_range = (4.0, 45.0)
            loc_od = (15.5, 1.5)
            plot_patch = False

        fig = None
        if ax is None:
                # No axes object given: create
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(1, figsize=(10, 8))

        # Matplotlib<2 compatibility
        if hasattr(ax, 'set_facecolor'):
            ax.set_facecolor('black')
        elif hasattr(ax, 'set_axis_bgcolor'):
            ax.set_axis_bgcolor('black')

        # Draw axon pathways: Need to regenerate because parfor returns axons
        # out of orders - random sample doesn't look neat. Use the same
        # settings as in the OFL, but with a smaller `n_axons`:
        phi = np.linspace(*phi_range, num=n_axons)
        func_kwargs = {'n_rho': n_rho, 'loc_od': loc_od,
                       'rho_range': rho_range, 'eye': self.implant.eye}
        axon_bundles = utils.parfor(retina.jansonius2009, phi,
                                    func_kwargs=func_kwargs,
                                    engine=self.engine, n_jobs=self.n_jobs,
                                    scheduler=self.scheduler)
        for bundle in axon_bundles:
            ax.plot(retina.dva2ret(bundle[:, 0]), retina.dva2ret(bundle[:, 1]),
                    c=(0.5, 1.0, 0.5))

        # Draw in the the retinal patch we're simulating.
        # This defines the size of our "percept" image below.
        if plot_patch:
            from matplotlib import patches
            patch = patches.Rectangle((x_range[0], y_range[0]),
                                      np.diff(x_range), np.diff(y_range),
                                      alpha=0.7)
            ax.add_patch(patch)

        # Highlight location of stimulated electrodes
        if stim is not None:
            for key in stim:
                el = self.implant[key]
                if el is not None:
                    ax.plot(el.x_center, el.y_center, 'oy',
                            markersize=np.sqrt(el.radius) * 2)

        # Plot all electrodes and their label
        for e in self.implant.electrodes:
            if annotate:
                ax.text(e.x_center + 100, e.y_center + 50, e.name,
                        color='white', size='x-large')
            ax.plot(e.x_center, e.y_center, 'ow', markersize=np.sqrt(e.radius))

        # Plot the location of the array's tack and annotate it
        if self.implant.tack:
            tx, ty = self.implant.tack
            ax.plot(tx, ty, 'ow')
            if annotate:
                if upside_down:
                    offset = 100
                else:
                    offset = -100
                ax.text(tx, ty + offset, 'tack',
                        horizontalalignment='center',
                        verticalalignment='top',
                        color='white', size='large')

        xmin, xmax, ymin, ymax = retina.dva2ret([-20, 20, -15, 15])
        ax.set_aspect('equal')
        ax.set_xlim(xmin, xmax)
        ax.set_xlabel('x (microns)')
        ax.set_ylim(ymin, ymax)
        ax.set_ylabel('y (microns)')
        eyestr = {'LE': 'left', 'RE': 'right'}
        ax.set_title('%s in %s eye' % (self.implant, eyestr[self.implant.eye]))
        ax.grid('off')

        if annotate:
            # Annotate the four retinal quadrants near the corners of the plot:
            # superior/inferior x temporal/nasal
            topbottom = ['top', 'bottom']
            temporalnasal = ['temporal', 'nasal']
            if upside_down:
                topbottom = ['bottom', 'top']
                temporalnasal = ['nasal', 'temporal']
            for yy, valign, si in zip([ymax, ymin], topbottom,
                                      ['superior', 'inferior']):
                for xx, halign, tn in zip([xmin, xmax], ['left', 'right'],
                                          temporalnasal):
                    ax.text(xx, yy, si + ' ' + tn,
                            color='black', fontsize=14,
                            horizontalalignment=halign,
                            verticalalignment=valign,
                            backgroundcolor=(1, 1, 1, 0.8))

        # Need to flip y axis to have upper half == upper visual field
        if upside_down:
            ax.invert_yaxis()

        return fig, ax


def get_brightest_frame(percept):
    """Returns the brightest frame of a percept

    This function returns the frame of a percept (brightness over time) that
    contains the brightest pixel.

    Parameters
    ----------
    percept : TimeSeries
        The brightness movie as a TimeSeries object.
    """
    _, frame = percept.max_frame()

    return frame



import numpy as np
import scipy.special as ss
import scipy.optimize as scpo
import numpy.testing as npt
import pytest
import logging

from .. import retina
from .. import implants
from .. import stimuli
from .. import utils


class LegacyNanduri2012(retina.Nanduri2012):
    """Preserve old implementation to make sure Cython model runs correctly"""

    def __init__(self, **kwargs):
        # Set default values of keyword arguments
        self.tau1 = 0.42 / 1000
        self.tau2 = 45.25 / 1000
        self.tau3 = 26.25 / 1000
        self.eps = 8.73
        self.asymptote = 14.0
        self.slope = 3.0
        self.shift = 16.0

        # Overwrite any given keyword arguments, print warning message (True)
        # if attempting to set an unrecognized keyword
        self.set_kwargs(True, **kwargs)

        _, self.gamma1 = utils.gamma(1, self.tau1, self.tsample)
        _, self.gamma2 = utils.gamma(1, self.tau2, self.tsample)
        _, self.gamma3 = utils.gamma(3, self.tau3, self.tsample)

    def model_cascade(self, in_arr, pt_list, layers, use_jit):
        """Nanduri model cascade

        Parameters
        ----------
        in_arr: array - like
            A 2D array specifying the effective current values
            at a particular spatial location(pixel); one value
            per retinal layer and electrode.
            Dimensions: <  # layers x #electrodes>
        pt_list: list
            List of pulse train 'data' containers.
            Dimensions: <  # electrodes x #time points>
        layers: list
            List of retinal layers to simulate.
            Choose from:
            - 'OFL': optic fiber layer
            - 'GCL': ganglion cell layer
        use_jit: bool
            If True, applies just - in-time(JIT) compilation to
            expensive computations for additional speed - up
            (requires Numba).
        """
        if 'INL' in layers:
            raise ValueError("The Nanduri2012 model does not support an inner "
                             "nuclear layer.")

        # Although the paper says to use cathodic-first, the code only
        # reproduces if we use what we now call anodic-first. So flip the sign
        # on the stimulus here:
        b1 = -self.calc_layer_current(in_arr, pt_list)

        # Fast response
        b2 = self.tsample * utils.conv(b1, self.gamma1, mode='full',
                                       method='sparse',
                                       use_jit=use_jit)[:b1.size]

        # Charge accumulation
        ca = self.tsample * np.cumsum(np.maximum(0, b1))
        ca = self.tsample * utils.conv(ca, self.gamma2, mode='full',
                                       method='fft')[:b1.size]
        b3 = np.maximum(0, b2 - self.eps * ca)

        # Stationary nonlinearity
        b3max = b3.max()
        sigmoid = ss.expit((b3max - self.shift) / self.slope)
        b4 = b3 / b3max * sigmoid * self.asymptote

        # Slow response
        b5 = self.tsample * utils.conv(b4, self.gamma3, mode='full',
                                       method='fft')[:b1.size]

        return utils.TimeSeries(self.tsample, b5)


def test_Grid():
    # Invalid calls
    with pytest.raises(ValueError):
        grid = retina.Grid(x_range=1, n_axons=1)
    with pytest.raises(ValueError):
        grid = retina.Grid(x_range=(1.0, 0.0), n_axons=1)
    with pytest.raises(ValueError):
        grid = retina.Grid(y_range=1, n_axons=1)
    with pytest.raises(ValueError):
        grid = retina.Grid(y_range=(1.0, 0.0), n_axons=1)
    for n_axons in [-1, 0]:
        with pytest.raises(ValueError):
            grid = retina.Grid(n_axons=n_axons)
    for n_rho in [-1, 0]:
        with pytest.raises(ValueError):
            retina.Grid(n_rho=n_rho)
    for lophi in [-200.0, 90.0, 200.0]:
        with pytest.raises(ValueError):
            retina.Grid(phi_range=(lophi, 85.0))
    for hiphi in [-200.0, 90.0, 200.0]:
        with pytest.raises(ValueError):
            retina.Grid(phi_range=(95.0, hiphi))

    # Verify size of axon bundles
    for n_axons in [3, 5, 10]:
        grid = retina.Grid(x_range=(0, 0), y_range=(0, 0), n_axons=n_axons,
                           save_data=False)
        # Number of axon bundles given by `n_axons`
        npt.assert_equal(len(grid.axon_bundles), n_axons)
        # Number of axons given by number of pixels on grid
        npt.assert_equal(len(grid.axons), 1)

    # Verify file


def test_BaseModel():
    # Cannot instantiate abstract class
    with pytest.raises(TypeError):
        tm = retina.BaseModel(0.01)

    # Child class must provide `model_cascade()`
    class Incomplete(retina.BaseModel):
        pass
    with pytest.raises(TypeError):
        tm = Incomplete()

    # A complete class
    class Complete(retina.BaseModel):

        def model_cascade(self, inval):
            return inval

    tm = Complete(tsample=0.1)
    npt.assert_equal(tm.tsample, 0.1)
    npt.assert_equal(tm.model_cascade(2.4), 2.4)


def test_TemporalModel():
    tsample = 0.01 / 1000

    # Assume 4 electrodes, each getting some stimulation
    pts = [stimuli.PulseTrain(tsample=tsample, dur=0.1)] * 4
    ptrain_data = [pt.data for pt in pts]

    # For each of these 4 electrodes, we have two effective current values:
    # one for the ganglion cell layer, one for the bipolar cell layer
    ecs_item = np.random.rand(2, 4)

    # What we want is the effective current per layer, for all electrodes
    # added up.
    # We do this for different layer configurations:
    for layers in [['INL'], ['GCL'], ['GCL', 'INL']]:
        ecm_by_hand = np.zeros((2, ptrain_data[0].size))
        if 'INL' in layers:
            for curr, pt in zip(ecs_item[0, :], ptrain_data):
                ecm_by_hand[0, :] += curr * pt
        if 'GCL' in layers:
            for curr, pt in zip(ecs_item[1, :], ptrain_data):
                ecm_by_hand[1, :] += curr * pt

        # And that should be the same as `calc_layer_current`:
        tm = retina.TemporalModel(tsample=tsample)
        ecm = tm.calc_layer_current(ecs_item, ptrain_data, layers)
        npt.assert_almost_equal(ecm, ecm_by_hand)
        tm.model_cascade(ecs_item, ptrain_data, layers, False)

    with pytest.raises(ValueError):
        tm.calc_layer_current(ecs_item, ptrain_data, ['unknown'])
    with pytest.raises(ValueError):
        tm.model_cascade(ecs_item, ptrain_data, ['unknown'], False)


def test_Nanduri2012():
    tsample = 0.01 / 1000
    tm = retina.Nanduri2012(tsample=tsample)

    # Assume 4 electrodes, each getting some stimulation
    pts = [stimuli.PulseTrain(tsample=tsample, dur=0.1)] * 4
    ptrain_data = [pt.data for pt in pts]

    # For each of these 4 electrodes, we have two effective current values:
    # one for the ganglion cell layer, one for the bipolar cell layer
    ecs_item = np.random.rand(2, 4)

    # ...and that should be the same as `calc_layer_current`:
    ecm_by_hand = np.sum(ecs_item[1, :, np.newaxis] * ptrain_data, axis=0)
    ecm = tm.calc_layer_current(ecs_item, ptrain_data)
    npt.assert_almost_equal(ecm, ecm_by_hand)

    # Running the model cascade:
    with pytest.raises(ValueError):
        tm.model_cascade(ecs_item, ptrain_data, ['GCL', 'INL'], False)
    with pytest.raises(ValueError):
        tm.model_cascade(ecs_item, ptrain_data, ['unknown'], False)

    # Regression test: Make sure Cython implementation reproduces legacy code
    tsample = 0.005 / 1000
    layers = ['GCL']
    use_jit = True
    stim = stimuli.PulseTrain(tsample, freq=20, amp=150, pulse_dur=0.45 / 1000,
                              dur=0.5)
    ecm = np.array([1, 1]).reshape((2, 1))
    legacy = LegacyNanduri2012(tsample=tsample)
    legacy_out = legacy.model_cascade(ecm, [stim.data], layers, use_jit)
    nanduri = retina.Nanduri2012(tsample=tsample)
    nanduri_out = nanduri.model_cascade(ecm, [stim.data], layers, use_jit)
    npt.assert_almost_equal(nanduri_out.data, legacy_out.data, decimal=2)
    npt.assert_almost_equal(nanduri_out.data[-1], legacy_out.data[-1],
                            decimal=2)


def test_Horsager2009_model_cascade():
    tsample = 0.01 / 1000
    tm = retina.Horsager2009(tsample=tsample)

    # Assume 4 electrodes, each getting some stimulation
    pts = [stimuli.PulseTrain(tsample=tsample, dur=0.1)] * 4
    ptrain_data = [pt.data for pt in pts]

    # For each of these 4 electrodes, we have two effective current values:
    # one for the ganglion cell layer, one for the bipolar cell layer
    ecs_item = np.random.rand(2, 4)

    # Run the model cascade:
    with pytest.raises(ValueError):
        tm.model_cascade(ecs_item, ptrain_data, ['GCL', 'INL'], False)
    with pytest.raises(ValueError):
        tm.model_cascade(ecs_item, ptrain_data, ['unknown'], False)
    tm.model_cascade(ecs_item, ptrain_data, ['GCL'], False)


def test_Horsager2009_calc_layer_current():
    tsample = 0.01 / 1000
    tm = retina.Horsager2009(tsample=tsample)

    # Assume 4 electrodes, each getting some stimulation
    pts = [stimuli.PulseTrain(tsample=tsample, dur=0.1)] * 4
    ptrain_data = [pt.data for pt in pts]

    # For each of these 4 electrodes, we have two effective current values:
    # one for the ganglion cell layer, one for the bipolar cell layer
    ecs_item = np.random.rand(2, 4)

    # Calulating layer current:
    # The Horsager model does not support INL, so it's just one layer:
    with pytest.raises(ValueError):
        tm.calc_layer_current(ecs_item, ptrain_data, ['GCL', 'INL'])
    with pytest.raises(ValueError):
        tm.calc_layer_current(ecs_item, ptrain_data, ['unknown'])

    # ...and that should be the same as `calc_layer_current`:
    ecm_by_hand = np.sum(ecs_item[1, :, np.newaxis] * ptrain_data, axis=0)
    ecm = tm.calc_layer_current(ecs_item, ptrain_data, ['GCL'])
    npt.assert_almost_equal(ecm, ecm_by_hand)


def test_Horsager2009():
    """Make sure the model can reproduce data from Horsager et al. (2009)

    Single-pulse data is taken from Fig.3, where the threshold current is
    reported for a given pulse duration and amplitude of a single pulse.
    We don't fit every data point to save some computation time, but make sure
    the model produces roughly the same values as reported in the paper for
    some data points.
    """

    def forward_pass(model, pdurs, amps):
        """Calculate model output based on a list of pulse durs and amps"""
        pdurs = np.array([pdurs]).ravel()
        amps = np.array([amps]).ravel()
        for pdur, amp in zip(pdurs, amps):
            in_arr = np.ones((2, 1))
            pt = stimuli.PulseTrain(model.tsample, amp=amp, freq=0.1,
                                    pulsetype='cathodicfirst',
                                    pulse_dur=pdur / 1000.0,
                                    interphase_dur=pdur / 1000.0)
            percept = model.model_cascade(in_arr, [pt.data], 'GCL', False)
            yield percept.data.max()

    def calc_error_amp(amp_pred, pdur, model):
        """Calculates the error in threshold current

        For a given data `pdur`, what is the `amp` needed for output `theta`?
        We're trying to find the current that produces output `theta`. Thus
        we calculate the error between output produced with `amp_pred` and
        `theta`.
        """
        theta_pred = list(forward_pass(model, pdur, amp_pred))[0]
        return np.log(np.maximum(1e-10, (theta_pred - model.theta) ** 2))

    def yield_fits(model, pdurs, amps):
        """Yields a threshold current by fitting the model to the data"""
        for pdur, amp in zip(pdurs, amps):
            yield scpo.fmin(calc_error_amp, amp, disp=0, args=(pdur, model))[0]

    # Data from Fig.3 in Horsager et al. (2009)
    pdurs = [0.07335, 0.21985, 0.52707, 3.96939]  # pulse duration in ms
    amps_true = [181.6, 64.7, 33.1, 14.7]  # threshold current in uA
    amps_expected = [201.9, 61.0, 29.2, 10.4]

    # Make sure our implementation comes close to ground-truth `amps`:
    # - Do the forward pass
    model = retina.Horsager2009(tsample=0.01 / 1000, tau1=0.42 / 1000,
                                tau2=45.25 / 1000, tau3=26.25 / 1000,
                                beta=3.43, epsilon=2.25, theta=110.3)
    amps_predicted = np.array(list(yield_fits(model, pdurs, amps_true)))

    # - Make sure predicted values still the same
    npt.assert_almost_equal(amps_expected, amps_predicted, decimal=0)


def test_Retina_Electrodes():
    logging.getLogger(__name__).info("test_Retina_Electrodes()")
    ssample = 1
    x_range = (-2, 2)
    y_range = (-3, 3)
    ret = retina.Grid(x_range=x_range, y_range=y_range, sampling=ssample,
                      save_data=False)
    npt.assert_equal(ret.gridx.shape, (int(np.diff(y_range) / ssample + 1),
                                       int(np.diff(x_range) / ssample + 1)))
    npt.assert_equal(ret.x_range, x_range)
    npt.assert_equal(ret.y_range, y_range)

    electrode1 = implants.Electrode('epiretinal', 1, 0, 0, 0)

    # Calculate current spread for all retinal layers
    retinal_layers = ['INL', 'OFL']
    cs = dict()
    ecs = dict()
    for layer in retinal_layers:
        cs[layer] = electrode1.current_spread(ret.gridx, ret.gridy,
                                              layer=layer)
        ecs[layer] = ret.current2effectivecurrent(cs[layer])

    electrode_array = implants.ElectrodeArray('epiretinal', [1, 1], [0, 1],
                                              [0, 1], [0, 1])
    npt.assert_equal(electrode1.x_center,
                     electrode_array.electrodes[0].x_center)
    npt.assert_equal(electrode1.y_center,
                     electrode_array.electrodes[0].y_center)
    npt.assert_equal(electrode1.radius, electrode_array.electrodes[0].radius)
    ecs_list, cs_list = ret.electrode_ecs(electrode_array)

    # Make sure cs_list has an entry for every layer
    npt.assert_equal(cs_list.shape[-2], len(retinal_layers))
    npt.assert_equal(ecs_list.shape[-2], len(retinal_layers))

    # Make sure manually generated current spreads match object
    for i, e in enumerate(retinal_layers):
        # last index: electrode, second-to-last: layer
        npt.assert_equal(cs[e], cs_list[..., i, 0])


def test_ret2dva():
    # Below 15mm eccentricity, relationship is linear with slope 3.731
    npt.assert_equal(retina.ret2dva(0.0), 0.0)
    for sign in [-1, 1]:
        for exp in [2, 3, 4]:
            ret = sign * 10 ** exp  # mm
            dva = 3.731 * sign * 10 ** (exp - 3)  # dva
            npt.assert_almost_equal(retina.ret2dva(ret), dva,
                                    decimal=3 - exp)  # adjust precision


def test_dva2ret():
    # Below 50deg eccentricity, relationship is linear with slope 0.268
    npt.assert_equal(retina.dva2ret(0.0), 0.0)
    for sign in [-1, 1]:
        for exp in [-2, -1, 0]:
            dva = sign * 10 ** exp  # deg
            ret = 0.268 * sign * 10 ** (exp + 3)  # mm
            npt.assert_almost_equal(retina.dva2ret(dva), ret,
                                    decimal=-exp)  # adjust precision


def test_jansonius2009():
    # With `rho` starting at 0, all axons should originate in the optic disc
    # center
    for phi0 in [-135.0, 66.0, 128.0]:
        for loc_od in [(15.0, 2.0), (-15.0, 2.0), (-4.2, -6.66)]:
            ax_pos = retina.jansonius2009(phi0, n_rho=100,
                                          rho_range=(0.0, 45.0),
                                          loc_od=loc_od)
            npt.assert_almost_equal(ax_pos[0, 0], loc_od[0])
            npt.assert_almost_equal(ax_pos[0, 1], loc_od[1])

    # These axons should all end at the meridian
    for sign in [-1.0, 1.0]:
        for phi0 in [110.0, 135.0, 160.0]:
            ax_pos = retina.jansonius2009(sign * phi0, n_rho=801,
                                          loc_od=(15, 2),
                                          rho_range=(0.0, 45.0))
            print(ax_pos[-1, :])
            npt.assert_almost_equal(ax_pos[-1, 1], 0.0, decimal=1)

    # `phi0` must be within [-180, 180]
    for phi0 in [-200.0, 181.0]:
        with pytest.raises(ValueError):
            retina.jansonius2009(phi0)

    # `n_rho` must be >= 1
    for n_rho in [-1, 0]:
        with pytest.raises(ValueError):
            retina.jansonius2009(0.0, n_rho=n_rho)

    # `rho_range` must have min <= max
    for lorho in [-200.0, 90.0]:
        with pytest.raises(ValueError):
            retina.jansonius2009(0.0, rho_range=(lorho, 45.0))
    for hirho in [-200.0, 40.0]:
        with pytest.raises(ValueError):
            retina.jansonius2009(0.0, rho_range=(45.0, hirho))

    # `eye` must be left or right
    for eye in ['L', 'r', 'left', 'right']:
        with pytest.raises(ValueError):
            retina.jansonius2009(0.0, eye=eye)

    # A single axon fiber with `phi0`=0 should return a single pixel location
    # that corresponds to the optic disc
    for eye in ['LE', 'RE']:
        for loc_od in [(15.5, 1.5), (7.0, 3.0), (-2.0, -2.0)]:
            single_fiber = retina.jansonius2009(0, n_rho=1, loc_od=loc_od,
                                                rho_range=(0, 0))
            npt.assert_equal(len(single_fiber), 1)
            npt.assert_almost_equal(single_fiber[0], loc_od)


def test_find_closest_axon():
    phi = np.linspace(-180.0, 180.0, 10)
    axon_bundles = utils.parfor(retina.jansonius2009, phi)
    for idx, ax in enumerate(axon_bundles):
        # Each axon bundle should be closest to itself
        closest = retina.find_closest_axon(ax[-1, :], axon_bundles)
        npt.assert_almost_equal(closest, ax[-1:0:-1, :])


def test_axon_dist_from_soma():
    # A small grid
    xg, yg = np.meshgrid([-1, 0, 1], [-1, 0, 1], indexing='xy')

    # When axon locations are snapped to the grid, a really short axon should
    # have zero distance to the soma:
    for x_soma in [-1.0, -0.2, 0.51]:
        axon = np.array([[i, i] for i in np.linspace(x_soma, x_soma + 0.01)])
        _, dist = retina.axon_dist_from_soma(axon, xg, yg)
        npt.assert_almost_equal(dist, 0.0)

    # On this simple grid, a diagonal axon should have dist [0, sqrt(2), 2]:
    for sign in [-1.0, 1.0]:
        for num in [10, 20, 50]:
            axon = np.array([[i, i] for i in np.linspace(sign, -sign, num)])
            _, dist = retina.axon_dist_from_soma(axon, xg, yg)
            npt.assert_almost_equal(dist, np.array([0.0, np.sqrt(2), 2.0]))

    # An axon that does not live near the grid should return infinite distance
    axon = np.array([[i, i] for i in np.linspace(1000.0, 1500.0)])
    _, dist = retina.axon_dist_from_soma(axon, xg, yg)
    npt.assert_equal(np.isinf(dist), True)


def test_axon_contribution():
    # Invalid calls
    for lmbda in [-1, 0]:
        with pytest.raises(ValueError):
            retina.axon_contribution([0], [0], decay_const=lmbda)
    for p in [-1, 0]:
        with pytest.raises(ValueError):
            retina.axon_contribution([0], [0], contribution_rule='mean',
                                     powermean_exp=p)
    with pytest.raises(ValueError):
        retina.axon_contribution([0], [0], contribution_rule='max',
                                 powermean_exp=1)
    with pytest.raises(ValueError):
        retina.axon_contribution([0], [0], sensitivity_rule='unknown')
    with pytest.raises(ValueError):
        retina.axon_contribution([0], [0], contribution_rule='unknown')

    # Make sure numbers are right for a simple setup:
    dist = np.arange(10)
    dist2 = (dist, dist)
    decay_const = 4.0
    for c_rule in ['max', 'mean', 'sum']:
        if c_rule == 'mean':
            powermean_exp = 1.0
        else:
            powermean_exp = None

        # If current spread == 1 everywhere, contribution given by `dist`
        _, contrib = retina.axon_contribution(
            dist2, np.ones_like(dist), sensitivity_rule='decay',
            decay_const=decay_const, contribution_rule=c_rule,
            powermean_exp=powermean_exp
        )
        sensitivity = np.exp(-dist / decay_const)
        method_to_call = getattr(np, c_rule)
        npt.assert_almost_equal(contrib, method_to_call(sensitivity))

    # No current spread in -> no axon contribution out
    for s_rule in ['decay', 'Jeng2011']:
        for c_rule in ['max', 'sum', 'mean']:
            for p in [1.0, 2.0, 3.0]:
                if c_rule == 'mean':
                    powermean_exp = p
                else:
                    powermean_exp = None
                _, contrib = retina.axon_contribution(
                    dist2, np.zeros_like(dist), sensitivity_rule=s_rule,
                    contribution_rule=c_rule, powermean_exp=powermean_exp
                )
                npt.assert_almost_equal(contrib, 0.0)


# deprecated
def test_make_axon_map():
    jan_x, jan_y = retina.jansonius(num_cells=10, num_samples=100)
    xg, yg = np.meshgrid(np.linspace(-100, 100, 21),
                         np.linspace(-100, 100, 21), indexing='xy')
    ax_id, ax_wt = retina.make_axon_map(xg, yg, jan_x, jan_y)



from __future__ import absolute_import, division, print_function
from .version import __version__  # noqa
import logging

# Disable Jupyter Notebook handlers
# https://github.com/ipython/ipython/issues/8282
logging.getLogger().handlers = []

# Set up root logger for debug file
formatstr = '%(asctime)s [%(name)s] [%(levelname)s] %(message)s'
logging.basicConfig(level=logging.DEBUG,
                    format=formatstr,
                    filename='debug.log',
                    filemode='w')

# Add streaming to console: Temporarily set level to ERROR so the imports
# don't trigger 'deprecated' warnings
console = logging.StreamHandler()
console.setLevel(logging.ERROR)
console.setFormatter(logging.Formatter(formatstr))
logging.getLogger(__name__).addHandler(console)

from .api import *
from . import retina
from . import implants
from . import stimuli
from . import files
from . import utils

# Reset the logging level to INFO
console.setLevel(logging.INFO)
logging.getLogger(__name__).info('Welcome to pulse2percept')



import numpy as np
import logging

from pulse2percept import utils

# Rather than trying to import these all over, try once and then remember
# by setting a flag.
try:
    import skimage
    import skimage.transform as sit
    import skimage.color as sic
    has_skimage = True
except (ImportError, AttributeError):
    # Might also raise "dict object has no attribute 'transform'"
    has_skimage = False

try:
    import skvideo
    import skvideo.io as svio
    skvideo._HAS_AVCONV
    skvideo._HAS_FFMPEG
    has_skvideo = True
except (ImportError, AttributeError):
    has_skvideo = False


def set_skvideo_path(ffmpeg_path=None, libav_path=None):
    """Sets the path to the FFMPEG and/or LibAV libraries.

    If scikit-video complains that either ffmpeg or libav cannot be found,
    you can set the path to the executables directly. On Unix, these binaries
    usually live in /usr/bin. On Windows, point to the directory that contains
    your ffmpeg.exe.

    Parameters
    ----------
    ffmpeg_path : str, optional, default: system's default path
        Path to ffmpeg library.
    libav_path : str, optional, default: system's default path
        Path to libav library.

    """
    if libav_path is not None:
        skvideo.setLibAVPath(libav_path)
    if ffmpeg_path is not None:
        skvideo.setFFmpegPath(ffmpeg_path)


def load_video_metadata(filename, ffmpeg_path=None, libav_path=None):
    """Returns a video files metadata

    This function loads the  metadata of a video file, which is returned
    as a dict. Among the available information are the width and height
    of the video, the aspect ratio, the frame rate, the bit rate, and the
    duration.

    Parameters
    ----------
    filename : str
        Video file name
    ffmpeg_path : str, optional, default: system's default path
        Path to ffmpeg library.
    libav_path : str, optional, default: system's default path
        Path to libav library.

    Returns
    -------
    metadata : dict
        A dictionary containing all metadata.
    """
    if not has_skvideo:
        raise ImportError("You do not have scikit-video installed. "
                          "You can install it via $ pip install sk-video.")

    # Set the path if necessary
    set_skvideo_path(ffmpeg_path, libav_path)

    metadata = skvideo.io.ffprobe(filename)
    if not metadata:
        raise OSError('File %s could not be found.' % filename)
    return metadata['video']


def load_video_framerate(filename, ffmpeg_path=None, libav_path=None):
    """Returns the video frame rate

    This function returns the frame rate of the video, as given by its
    metadata field '@r_frame_rate'.

    Parameters
    ----------
    filename : str
        Video file name
    ffmpeg_path : str, optional, default: system's default path
        Path to ffmpeg library.
    libav_path : str, optional, default: system's default path
        Path to libav library.

    Returns
    -------
    fps : float
        Video frame rate (frames per second).
    """
    if not has_skvideo:
        raise ImportError("You do not have scikit-video installed. "
                          "You can install it via $ pip install sk-video.")

    # Set the path if necessary
    set_skvideo_path(ffmpeg_path, libav_path)

    # Load all metadata
    metadata = load_video_metadata(filename)
    if '@r_frame_rate' not in metadata:
        raise AttributeError('Meta data does not contain field @r_frame_rate.')

    # Parse frame rate entry: It's a rational expression encoded as a string.
    # Hence, split by '/' and to the divison by hand.
    str_fps = metadata['@r_frame_rate'].split('/')
    return float(str_fps[0]) / float(str_fps[1])


def load_video(filename, as_timeseries=True, as_gray=False, ffmpeg_path=None,
               libav_path=None):
    """Loads a video from file.

    This function loads a video from file with the help of Scikit-Video, and
    returns the data either as a NumPy array (if `as_timeseries` is False)
    or as a ``p2p.utils.TimeSeries`` object (if `as_timeseries` is True).

    Parameters
    ----------
    filename : str
        Video file name
    as_timeseries: bool, optional, default: True
        If True, returns the data as a ``p2p.utils.TimeSeries`` object.
    as_gray : bool, optional, default: False
        If True, loads only the luminance channel of the video.
    ffmpeg_path : str, optional, default: system's default path
        Path to ffmpeg library.
    libav_path : str, optional, default: system's default path
        Path to libav library.

    Returns
    -------
    video : ndarray | p2p2.utils.TimeSeries
        If `as_timeseries` is False, returns video data according to the
        Scikit-Video standard; that is, an ndarray of dimension (T, M, N, C),
        (T, M, N), (M, N, C), or (M, N), where T is the number of frames,
        M is the height, N is the width, and C is the number of channels (will
        be either 1 for grayscale or 3 for RGB).

        If `as_timeseries` is True, returns video data as a TimeSeries object
        of dimension (M, N, C), (M, N, T), (M, N, C), or (M, N).
        The sampling rate corresponds to 1 / frame rate.

    Examples
    --------
    Load a video as a ``p2p.utils.TimeSeries`` object:

    >>> from skvideo import datasets
    >>> video = load_video(datasets.bikes())
    >>> video.tsample
    0.04
    >>> video.shape
    (272, 640, 3, 250)

    Load a video as a NumPy ndarray:

    >>> from skvideo import datasets
    >>> video = load_video(datasets.bikes(), as_timeseries=False)
    >>> video.shape
    (250, 272, 640, 3)

    Load a video as a NumPy ndarray and convert to grayscale:

    >>> from skvideo import datasets
    >>> video = load_video(datasets.bikes(), as_timeseries=False, as_gray=True)
    >>> video.shape
    (250, 272, 640, 1)

    """
    if not has_skvideo:
        raise ImportError("You do not have scikit-video installed. "
                          "You can install it via $ pip install sk-video.")

    # Set the path if necessary
    set_skvideo_path(ffmpeg_path, libav_path)

    if skvideo._HAS_FFMPEG:
        backend = 'ffmpeg'
    else:
        backend = 'libav'
    video = svio.vread(filename, as_grey=as_gray, backend=backend)
    logging.getLogger(__name__).info("Loaded video from file '%s'." % filename)
    d_s = "Loaded video has shape (T, M, N, C) = " + str(video.shape)
    logging.getLogger(__name__).debug(d_s)

    if as_timeseries:
        # TimeSeries has the time as the last dimensions: re-order dimensions,
        # then squeeze out singleton dimensions
        axes = np.roll(range(video.ndim), -1)
        video = np.squeeze(np.transpose(video, axes=axes))
        fps = load_video_framerate(filename)
        d_s = "Reshaped video to shape (M, N, C, T) = " + str(video.shape)
        logging.getLogger(__name__).debug(d_s)
        return utils.TimeSeries(1.0 / fps, video)
    else:
        # Return as ndarray
        return video


def load_video_generator(filename, ffmpeg_path=None, libav_path=None):
    """Returns a generator that can load a video from file frame-by-frame.

    This function returns a generator `reader` that can load a video from a
    file frame-by-frame. Every call to `reader.nextFrame()` will return a
    single frame of the video as a NumPy array with dimensions (M, N) or
    (M, N, C), where M is the height, N is the width, and C is the number of
    channels (will be either 1 for grayscale or 3 for RGB).

    Parameters
    ----------
    filename : str
        Video file name
    ffmpeg_path : str, optional, default: system's default path
        Path to ffmpeg library.
    libav_path : str, optional, default: system's default path
        Path to libav library.

    Returns
    -------
    reader : skvideo.io.FFmpegReader | skvideo.io.LibAVReader
        A Scikit-Video reader object

    Examples
    --------
    >>> from skvideo import datasets
    >>> reader = load_video_generator(datasets.bikes())
    >>> for frame in reader.nextFrame():
    ...    pass

    """
    if not has_skvideo:
        raise ImportError("You do not have scikit-video installed. "
                          "You can install it via $ pip install sk-video.")

    # Set the path if necessary
    set_skvideo_path(ffmpeg_path, libav_path)

    # Try loading
    if skvideo._HAS_FFMPEG:
        reader = svio.FFmpegReader(filename)
    elif skvideo._HAS_AVCONV:
        reader = svio.LibAVReader(filename)
    else:
        raise ImportError("You have neither ffmpeg nor libav (which comes "
                          "with avprobe) installed.")

    logging.getLogger(__name__).info("Loaded video from file '%s'." % filename)
    return reader


def save_video(data, filename, width=None, height=None, fps=30,
               ffmpeg_path=None, libav_path=None):
    """Saves a video to file.

    This function stores a NumPy ndarray to file using Scikit-Video.

    Parameters
    ----------
    data : ndarray | p2p.utils.TimeSeries
        Video data as a NumPy ndarray must have dimension (T, M, N, C),
        (T, M, N), (M, N, C), or (M, N), where T is the number of frames,
        M is the height, N is the width, and C is the number of channels (must
        be either 1 for grayscale or 3 for RGB).

        Video data as a TimeSeries object must have dimension (M, N, C, T) or
        (M, N, T). The sampling step will be used as the video's frame rate.
    filename : str
        Video file name.
    width : int, optional
        Desired width of the movie.
        Default: Automatically determined based on `height` (without changing
        the aspect ratio). If `height` is not given, the percept's original
        width is used.
    height : int, optional
        Desired height of the movie.
        Default: Automatically determined based on `width` (without changing
        the aspect ratio). If `width` is not given, the percept's original
        height is used.
    fps : int, optional, default: 30
        Desired frame rate of the video (frames per second).
    ffmpeg_path : str, optional, default: system's default path
        Path to ffmpeg library.
    libav_path : str, optional, default: system's default path
        Path to libav library.

    Notes
    -----
    To distinguish between 3-D inputs of shape (T, M, N) and (M, N, C), the
    last dimension is checked: If it is small (<= 4), it is most likely a
    color channel - hence the input is interpreted as (M, N, C).
    Else it is interpreted as (T, M, N).

    """
    if not has_skvideo:
        raise ImportError("You do not have scikit-video installed. "
                          "You can install it via $ pip install sk-video.")
    if not has_skimage:
        raise ImportError("You do not have scikit-image installed. "
                          "You can install it via $ pip install scikit-image.")

    is_ndarray = is_timeseries = False
    if isinstance(data, np.ndarray):
        is_ndarray = True
    elif isinstance(data, utils.TimeSeries):
        is_timeseries = True
    else:
        raise TypeError('Data to be saved must be either a NumPy ndarray '
                        'or a ``p2p.utils.TimeSeries`` object')

    # Set the path if necessary, then choose backend
    set_skvideo_path(ffmpeg_path, libav_path)
    if skvideo._HAS_FFMPEG:
        backend = 'ffmpeg'
    else:
        backend = 'libav'

    if is_ndarray:
        # Use Scikit-Video utility to interpret dimensions and convert to
        # (T, M, N, C)
        data = skvideo.utils.vshape(data)
        oldheight = data.shape[1]
        oldwidth = data.shape[2]
        length = data.shape[0]
    elif is_timeseries:
        # Resample the percept to the given frame rate
        new_tsample = 1.0 / float(fps)
        d_s = 'old: tsample=%f' % data.tsample
        d_s += ', shape=' + str(data.shape)
        logging.getLogger(__name__).debug(d_s)
        data = data.resample(new_tsample)
        d_s = 'new: tsample=%f' % new_tsample
        d_s += ', shape=' + str(data.shape)
        logging.getLogger(__name__).debug(d_s)
        oldheight = data.shape[0]
        oldwidth = data.shape[1]
        length = data.shape[-1]

    # Calculate the desired height and width
    if not height and not width:
        height = oldheight
        width = oldwidth
    elif height and not width:
        width = int(height * 1.0 / oldheight * oldwidth)
    elif width and not height:
        height = int(width * 1.0 / oldwidth * oldheight)
    d_s = "Video scaled to (M, N, T) = (%d, %d, %d)." % (height, width, length)
    logging.getLogger(__name__).debug(d_s)

    # Reshape and scale the data
    savedata = np.zeros((length, height, width, 3), dtype=np.float32)
    for i in range(length):
        if is_ndarray:
            frame = skimage.img_as_float(data[i, ...])
        elif is_timeseries:
            frame = data.data[..., i] / float(data.data.max())
            frame = skimage.img_as_float(frame)

        # resize wants the data to be between 0 and 1
        frame = sic.gray2rgb(sit.resize(frame, (height, width),
                                        mode='reflect'))
        savedata[i, ...] = frame * 255.0

    # Set the input frame rate and frame size
    inputdict = {}
    inputdict['-r'] = str(int(fps))
    inputdict['-s'] = '%dx%d' % (width, height)

    # Set the output frame rate
    outputdict = {}
    outputdict['-r'] = str(int(fps))

    # Save data to file
    svio.vwrite(filename, savedata, inputdict=inputdict, outputdict=outputdict,
                backend=backend)
    logging.getLogger(__name__).info("Saved video to file '%s'." % filename)


def save_video_sidebyside(videofile, percept, savefile, fps=30,
                          ffmpeg_path=None, libav_path=None):
    """Saves both an input video and the percept to file, side-by-side.

    This function creates a new video from an input video file and a
    ``p2p.utils.TimeSeries`` object, assuming they correspond to model
    input and model output, and plots them side-by-side.
    Both input video and percept are resampled according to `fps`.
    The percept is resized to match the height of the input video.

    Parameters
    ----------
    videofile : str
        File name of input video.
    percept : p2p.utils.TimeSeries
        A TimeSeries object with dimension (M, N, C, T) or (M, N, T), where
        T is the number of frames, M is the height, N is the width, and C is
        the number of channels.
    savefile : str
        File name of output video.
    fps : int, optional, default: 30
        Desired frame rate of output video.
    ffmpeg_path : str, optional, default: system's default path
        Path to ffmpeg library.
    libav_path : str, optional, default: system's default path
        Path to libav library.

    """
    if not has_skvideo:
        raise ImportError("You do not have scikit-video installed. "
                          "You can install it via $ pip install sk-video.")
    if not has_skimage:
        raise ImportError("You do not have scikit-image installed. "
                          "You can install it via $ pip install scikit-image.")

    if not isinstance(percept, utils.TimeSeries):
        raise TypeError("`percept` must be of type p2p.utils.TimeSeries.")

    # Set the path if necessary
    set_skvideo_path(ffmpeg_path, libav_path)

    # Load video from file
    video = load_video(videofile, as_timeseries=True, ffmpeg_path=ffmpeg_path,
                       libav_path=libav_path)

    # Re-sample time series to new frame rate
    new_tsample = 1.0 / float(fps)
    video = video.resample(new_tsample)
    percept = percept.resample(new_tsample)

    # After re-sampling, both video and percept should have about the same
    # length (up to some rounding error)
    combined_len = np.minimum(video.shape[-1], percept.shape[-1])
    if np.abs(video.shape[-1] - percept.shape[-1]) > 5:
        raise ValueError('Could not resample percept, len(video)='
                         '%d != len(percept)=%d.' % (video.shape[-1],
                                                     percept.shape[-1]))

    # Up-scale percept to fit video dimensions
    pheight = video.shape[0]
    pwidth = int(percept.shape[1] / float(percept.shape[0]) * video.shape[0])

    # Show the two side-by-side
    combined = np.zeros((combined_len, pheight, video.shape[1] + pwidth, 3))
    combined = combined.astype(np.float32)
    for i in range(combined_len):
        vframe = skimage.img_as_float(video.data[..., i])
        pframe = percept.data[..., i] / percept.data.max()
        pframe = skimage.img_as_float(pframe)
        pframe = sic.gray2rgb(sit.resize(pframe, (pheight, pwidth),
                                         mode='reflect'))
        combined[i, ...] = np.concatenate((vframe, pframe), axis=1)

    save_video(combined, savefile, fps=fps, ffmpeg_path=ffmpeg_path,
               libav_path=libav_path)


def find_files_like(datapath, pattern):
    """Finds files in a folder whose name matches a pattern

    This function looks for files in folder `datapath` that match a regular
    expression `pattern`.

    Parameters
    ----------
    datapath : str
        Path to search
    pattern : str
        A valid regular expression pattern

    Examples
    --------
    # Find all '.npz' files in parent dir
    >>> files = find_files_like('..', '.*\.npz$')
    """
    # No need to import these at module level
    from os import listdir
    import re

    # Traverse file list and look for `pattern`
    filenames = []
    pattern = re.compile(pattern)
    for file in listdir(datapath):
        if pattern.search(file):
            filenames.append(file)

    return filenames



import os
import numpy as np
from setuptools import find_packages, setup
from setuptools.extension import Extension
from Cython.Build import cythonize

extensions = [
    Extension('pulse2percept.fast_retina', ['pulse2percept/fast_retina.pyx'],
              include_dirs=[np.get_include()],
              extra_compile_args=['-O3'])
]

# Get version and release info, which is all stored in pulse2percept/version.py
ver_file = os.path.join('pulse2percept', 'version.py')
with open(ver_file) as f:
    exec(f.read())

opts = dict(name=NAME,
            maintainer=MAINTAINER,
            maintainer_email=MAINTAINER_EMAIL,
            description=DESCRIPTION,
            long_description=LONG_DESCRIPTION,
            url=URL,
            download_url=DOWNLOAD_URL,
            license=LICENSE,
            classifiers=CLASSIFIERS,
            author=AUTHOR,
            author_email=AUTHOR_EMAIL,
            platforms=PLATFORMS,
            version=VERSION,
            packages=find_packages(),
            ext_modules=cythonize(extensions),
            install_requires=REQUIRES)


if __name__ == '__main__':
    setup(**opts)



import numpy as np
import numpy.testing as npt
import pytest
import copy

from .. import utils

try:
    # Python 3
    from unittest import mock
except ImportError:
    # Python 2
    import mock

try:
    # Python 3
    from imp import reload
except ImportError:
    pass


@utils.deprecated(behavior='raise')
def raise_deprecated():
    pass


def test_deprecated():
    with pytest.raises(RuntimeError):
        raise_deprecated()


def test_TimeSeries():
    max_val = 2.0
    max_idx = 156
    data_orig = np.random.rand(10, 10, 1000)
    data_orig[4, 4, max_idx] = max_val
    ts = utils.TimeSeries(1.0, data_orig)

    # Make sure function returns largest element
    tmax, vmax = ts.max()
    npt.assert_equal(tmax, max_idx)
    npt.assert_equal(vmax, max_val)

    # Make sure function returns largest frame
    tmax, fmax = ts.max_frame()
    npt.assert_equal(tmax, max_idx)
    npt.assert_equal(fmax.data, data_orig[:, :, max_idx])

    # Make sure getitem works
    npt.assert_equal(isinstance(ts[3], utils.TimeSeries), True)
    npt.assert_equal(ts[3].data, ts.data[3])


def test_TimeSeries_resample():
    max_val = 2.0
    max_idx = 156
    data_orig = np.random.rand(10, 10, 1000)
    data_orig[4, 4, max_idx] = max_val
    ts = utils.TimeSeries(1.0, data_orig)
    tmax, vmax = ts.max()

    # Resampling with same sampling step shouldn't change anything
    ts_new = ts.resample(ts.tsample)
    npt.assert_equal(ts_new.shape, ts.shape)
    npt.assert_equal(ts_new.duration, ts.duration)

    # Make sure resampling works
    tsample_new = 4
    ts_new = ts.resample(tsample_new)
    npt.assert_equal(ts_new.tsample, tsample_new)
    npt.assert_equal(ts_new.data.shape[-1], ts.data.shape[-1] / tsample_new)
    npt.assert_equal(ts_new.duration, ts.duration)
    tmax_new, vmax_new = ts_new.max()
    npt.assert_equal(tmax_new, tmax)
    npt.assert_equal(vmax_new, vmax)

    # Make sure resampling leaves old data unaffected (deep copy)
    ts_new.data[0, 0, 0] = max_val * 2.0
    tmax_new, vmax_new = ts_new.max()
    npt.assert_equal(tmax_new, 0)
    npt.assert_equal(vmax_new, max_val * 2.0)
    tmax, vmax = ts.max()
    npt.assert_equal(tmax, max_idx)
    npt.assert_equal(vmax, max_val)


def test_TimeSeries_append():
    max_val = 2.0
    max_idx = 156
    data_orig = np.random.rand(10, 10, 1000)
    data_orig[4, 4, max_idx] = max_val
    ts_orig = utils.TimeSeries(1.0, data_orig)

    # Make sure adding two TimeSeries objects works:
    # Must have the right type and size
    ts = copy.deepcopy(ts_orig)
    with pytest.raises(TypeError):
        ts.append(4.0)
    with pytest.raises(ValueError):
        ts_wrong_size = utils.TimeSeries(1.0, np.ones((2, 2)))
        ts.append(ts_wrong_size)

    # Adding messes only with the last dimension of the array
    ts = copy.deepcopy(ts_orig)
    ts.append(ts)
    npt.assert_equal(ts.shape[:-1], ts_orig.shape[:-1])
    npt.assert_equal(ts.shape[-1], ts_orig.shape[-1] * 2)

    # If necessary, the second pulse train is resampled to the first
    ts = copy.deepcopy(ts_orig)
    tsample_new = 2.0
    ts_new = ts.resample(tsample_new)
    ts.append(ts_new)
    npt.assert_equal(ts.shape[:-1], ts_orig.shape[:-1])
    npt.assert_equal(ts.shape[-1], ts_orig.shape[-1] * 2)
    ts_add = copy.deepcopy(ts_new)
    ts_add.append(ts_orig)
    npt.assert_equal(ts_add.shape[:-1], ts_new.shape[:-1])
    npt.assert_equal(ts_add.shape[-1], ts_new.shape[-1] * 2)


def test_sparseconv():
    # time vector for stimulus (long)
    maxT = .5  # seconds
    nt = 100000
    t = np.linspace(0, maxT, nt)

    # stimulus (10 Hz anondic and cathodic pulse train)
    stim = np.zeros(nt)
    stim[0:nt:10000] = 1
    stim[100:nt:1000] = -1

    # time vector for impulse response (shorter)
    tt = t[t < .1]

    # impulse reponse (kernel)
    G = np.exp(-tt / .005)

    # make sure sparseconv returns the same result as np.convolve
    # for all modes
    modes = ["full", "valid", "same"]
    for mode in modes:
        # np.convolve
        conv = np.convolve(stim, G, mode=mode)

        # utils.sparseconv
        sparse_conv = utils.sparseconv(stim, G, mode=mode, use_jit=False)

        npt.assert_equal(conv.shape, sparse_conv.shape)
        npt.assert_almost_equal(conv, sparse_conv)

    with pytest.raises(ValueError):
        utils.sparseconv(G, stim, mode='invalid')


def test_conv():
    reload(utils)
    # time vector for stimulus (long)
    stim_dur = 0.5  # seconds
    tsample = 0.001 / 1000
    t = np.arange(0, stim_dur, tsample)

    # stimulus (10 Hz anondic and cathodic pulse train)
    stim = np.zeros_like(t)
    stim[::1000] = 1
    stim[100::1000] = -1

    # kernel
    _, gg = utils.gamma(1, 0.005, tsample)

    # make sure conv returns the same result as
    # make sure sparseconv returns the same result as np.convolve
    # for all modes
    methods = ["fft", "sparse"]
    modes = ["full", "valid", "same"]
    for mode in modes:
        # np.convolve
        npconv = np.convolve(stim, gg, mode=mode)

        for method in methods:
            conv = utils.conv(stim, gg, mode=mode, method=method)

            npt.assert_equal(conv.shape, npconv.shape)
            npt.assert_almost_equal(conv, npconv)

    with pytest.raises(ValueError):
        utils.conv(gg, stim, mode="invalid")
    with pytest.raises(ValueError):
        utils.conv(gg, stim, method="invalid")

    with mock.patch.dict("sys.modules", {"numba": {}}):
        with pytest.raises(ImportError):
            reload(utils)
            utils.conv(stim, gg, mode='full', method='sparse', use_jit=True)


def power_it(num, n=2):
    return num ** n


def test_parfor():
    my_array = np.arange(100).reshape(10, 10)
    i, j = np.random.randint(0, 9, 2)
    my_list = list(my_array.ravel())

    expected_00 = power_it(my_array[0, 0])
    expected_ij = power_it(my_array[i, j])

    with pytest.raises(ValueError):
        utils.parfor(power_it, my_list, engine='unknown')
    with pytest.raises(ValueError):
        utils.parfor(power_it, my_list, engine='dask', scheduler='unknown')

    for engine in ['serial', 'joblib', 'dask']:
        for scheduler in ['threading', 'multiprocessing']:
            # `backend` only relevant for dask, will be ignored for others
            # and should thus still give the right result
            calculated_00 = utils.parfor(power_it, my_list, engine=engine,
                                         scheduler=scheduler,
                                         out_shape=my_array.shape)[0, 0]
            calculated_ij = utils.parfor(power_it, my_list, engine=engine,
                                         scheduler=scheduler,
                                         out_shape=my_array.shape)[i, j]

        npt.assert_equal(expected_00, calculated_00)
        npt.assert_equal(expected_ij, calculated_ij)

        if engine == 'serial':
            continue

        with mock.patch.dict("sys.modules", {engine: {}}):
            reload(utils)
            with pytest.raises(ImportError):
                utils.parfor(power_it, my_list, engine=engine,
                             out_shape=my_array.shape)[0, 0]
        reload(utils)


def test_gamma():
    tsample = 0.005 / 1000

    with pytest.raises(ValueError):
        t, g = utils.gamma(0, 0.1, tsample)
    with pytest.raises(ValueError):
        t, g = utils.gamma(2, -0.1, tsample)
    with pytest.raises(ValueError):
        t, g = utils.gamma(2, 0.1, -tsample)

    for tau in [0.001, 0.01, 0.1]:
        for n in [1, 2, 5]:
            t, g = utils.gamma(n, tau, tsample)
            npt.assert_equal(np.arange(0, t[-1] + tsample / 2.0, tsample), t)
            if n > 1:
                npt.assert_equal(g[0], 0.0)

            # Make sure area under the curve is normalized
            npt.assert_almost_equal(np.trapz(np.abs(g), dx=tsample), 1.0,
                                    decimal=2)

            # Make sure peak sits correctly
            npt.assert_almost_equal(g.argmax() * tsample, tau * (n - 1))


def test_traverse_randomly():
    # Generate a list of integers
    sequence = np.arange(100).tolist()

    # Iterate the sequence in random order
    shuffled_idx = []
    shuffled_val = []
    for idx, value in enumerate(utils.traverse_randomly(sequence)):
        shuffled_idx.append(idx)
        shuffled_val.append(value)

    # Make sure every element was visited once
    npt.assert_equal(shuffled_idx, np.arange(len(sequence)).tolist())
    npt.assert_equal(shuffled_val.sort(), sequence.sort())
